//: Playground - noun: a place where people can play

import UIKit
import XCPlayground

class Factory<T: Decodable> {
    let typeKeyPath: String
    var subtypes = [String: T.Type]()
    
    init(typeKeyPath: String) {
        self.typeKeyPath = typeKeyPath
    }
    
    func register(typeKey: String, asType: T.Type) {
        subtypes[typeKey] = asType
    }
    
    func make(json: JSON) -> T? {
        if let type = json.valueForKeyPath(typeKeyPath) as? String,
            subtype = subtypes[type] {
                return subtype.init(json: json)
        }
        return nil
    }
}

protocol FactoryDecodable: Decodable {
    associatedtype Output: Decodable
    static var factory: Factory<Output> { get }
}

class Shape: FactoryDecodable {
    static var factory = Factory<Shape>(typeKeyPath: "type")
    
    var area: Double { return 0 }
    
    required init?(json: JSON) {
        // base class does nothing
    }
    
}

class Square: Shape {
    let side: Double
    override var area: Double { return side * side }
    
    required init?(json: JSON) {
        if let side: Double = "side" <~~ json {
            self.side = side
            super.init(json: json)
        } else {
            return nil
        }
    }
}

class Circle: Shape {
    let radius: Double
    override var area: Double { return M_PI * radius * radius }
    
    required init?(json: JSON) {
        if let radius: Double = "radius" <~~ json {
            self.radius = radius
            super.init(json: json)
        } else {
            return nil
        }
    }
}

Shape.factory.register("square", asType: Square.self)
Shape.factory.register("circle", asType: Circle.self)

extension Decoder {
    
    static func decodeWithFactory<T: FactoryDecodable>(key: String, json: JSON) -> T? {
        if let keyed = json.valueForKeyPath(key) as? JSON {
            return T.factory.make(keyed) as? T
        }
        return nil
    }
    
    static func decodeArrayWithFactory<T: FactoryDecodable>(key: String, json: JSON) -> [T]? {
        guard let array = json.valueForKeyPath(key) as? [JSON] else { return nil }
        return array.flatMap { T.factory.make($0) as? T }
    }
    
    static func decodeDictionaryWithFactory<T: FactoryDecodable>(key: String, json: JSON) -> [String: T]? {
        guard let dict = json.valueForKeyPath(key) as? [String: JSON] else { return nil }
        var made = [String: T]()
        for (key, value) in dict {
            if let element = T.factory.make(value) as? T {
                made[key] = element
            }
        }
        return made
    }
}

infix operator <~*~ { associativity left precedence 150 }


func <~*~ <T: FactoryDecodable>(key: String, json: JSON) -> T? {
    return Decoder.decodeWithFactory(key, json: json)
}

func <~*~ <T: FactoryDecodable>(key: String, json: JSON) -> [T]? {
    return Decoder.decodeArrayWithFactory(key, json: json)
}

func <~*~ <T: FactoryDecodable>(key: String, json: JSON) -> [String: T]? {
    return Decoder.decodeDictionaryWithFactory(key, json: json)
}

class ShapeHolder: Decodable {
    let shapes: [Shape]
    let byKey: [String: Shape]
    
    required init?(json: JSON) {
        shapes = "shapes" <~*~ json ?? []
        byKey = "byKey" <~*~ json ?? [:]
    }
}



let filePath = NSBundle.mainBundle().pathForResource("shapes", ofType:"json")
let data = try! NSData(contentsOfFile:filePath!,
    options: NSDataReadingOptions.DataReadingUncached)
let json = try! NSJSONSerialization.JSONObjectWithData(data,
    options: NSJSONReadingOptions()) as! JSON

if let holder = ShapeHolder(json: json) {
    print("array: \(holder.shapes)")
    print("dictionary: \(holder.byKey)")
    
    print("arrayArea: \(holder.shapes.reduce(0, combine: { $0 + $1.area })) = 9 + π")
    print("dictioaryArea: \(holder.byKey.values.reduce(0, combine: { $0 + $1.area })) = 25 + 100π")
}

